// SPDX-License-Identifier: GPL-2.0-or-later

/*
* Copyright (C) 2025 Alibaba Inc
*/

#include <dirent.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utmp.h>
#include <zlib.h>

#include "bitfield.h"
#include "ras-events.h"
#include "ras-erst.h"
#include "ras-logger.h"
#include "ras-mce-handler.h"
#include "ras-record.h"
#include "ras-report.h"
#include "types.h"

struct apei_regex {
	regex_t hdr;
	regex_t severity;
	regex_t error;
	regex_t fru;
	regex_t type;

	regex_t addr;
	regex_t loc;
	regex_t mem_type;
	regex_t mem_status;

	regex_t port_type;
	regex_t port;
	regex_t id;
	regex_t status;
	regex_t aer_sev;
	regex_t tlp_hdr;

	regex_t cpu_id;

	regex_t midr;
	regex_t mpidr;
};

enum {
	APEI_NONE,
	APEI_CPU,
	APEI_MEM,
	APEI_PCIE,
	APEI_ARM,
};

struct apei {
	int id;
	int sev;
	int err_id;
	char *fru;
	int type;
	time_t time;
	union {
		struct {
			uint64_t addr;
			char *loc;
			char *status;
			char *type;
		} mem;
		struct {
			int port_type;
			char *port;
			char *vendor_id;
			char *device_id;
			char *status;
			char *mask;
			char *sev;
			char *tlp_hdr;
		} pcie;
		struct {
			char *cpu_id;
		} cpu;
		struct {
			char *midr;
			char *mpidr;
		} arm;
	};
};

time_t last_reboot_time;

static void get_last_reboot_time(void)
{
	struct utmp record;
	int fd;
	int reboots_found = 0;
	time_t reboot_times;

	fd = open("/var/log/wtmp", O_RDONLY);
	if (fd == -1) {
		log(ALL, LOG_ERR, "Error opening wtmp file");
		return;
	}

	if (lseek(fd, -1 * sizeof(struct utmp), SEEK_END) == -1) {
		perror("Error seeking in wtmp file");
		close(fd);
		return;
	}

	while (reboots_found < LAST_REBOOT_INDEX) {
		if (read(fd, &record, sizeof(struct utmp)) != sizeof(struct utmp)) {
			perror("Error reading wtmp file");
			close(fd);
			return;
		}

		if (strncmp(record.ut_line, "~", 1) == 0) {
			if (strncmp(record.ut_user, "reboot", 6) == 0) {
				reboot_times = record.ut_tv.tv_sec;
				reboots_found++;
			}
		}

		if (lseek(fd, -2 * sizeof(struct utmp), SEEK_CUR) == -1) {
			reboot_times = 0;
			break;
		}
	}

	close(fd);

	last_reboot_time = reboot_times;

	return;
}

#define DMESG_ERST_PREFIX "dmesg-erst"
#define DMESG_ERST_SUFFIX "enc.z"

#define APEI_HEADER	".*\\[(.*).[0-9]+\\] \\{([0-9]+)\\}\\[Hardware Error\\]: Hardware error from APEI Generic Hardware Error Source:.*"
#define APEI_SEVERITY	".*\\{([0-9]+)\\}\\[Hardware Error\\]: event severity: (.*)"
#define APEI_ERROR	".*\\{([0-9]+)\\}\\[Hardware Error\\]:  Error ([0-9]+), type: (.*)"
#define APEI_MEM_FRU	".*\\{([0-9]+)\\}\\[Hardware Error\\]:  fru_text: (.*)"
#define APEI_TYPE	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   section_type: (.*)"

// MEM
#define APEI_MEM_ADDR	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   physical_address: (.*)"
#define APEI_MEM_LOC	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   (node:.*)"
#define APEI_MEM_TYPE	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   error_type: [0-9]+, (.*)"
#define APEI_MEM_STATUS	".*\\{([0-9]+)\\}\\[Hardware Error\\]:.*error_status: (.*) \\(.*\\)"

// PCIE
#define APEI_PORT_TYPE	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   port_type: ([0-9]+), (.*)"
#define APEI_PORT	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   device_id: (.*)"
#define APEI_ID		".*\\{([0-9]+)\\}\\[Hardware Error\\]:   vendor_id: (.*), device_id: (.*)"
#define APEI_STATUS	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   aer_uncor_status: (.*), aer_uncor_mask: (.*)"
#define APEI_AER_SEVE	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   aer_uncor_severity: (.*)"
#define APEI_TLP_HDR	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   TLP Header: (.*)"

#define APEI_CPU_ID	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   processor_id: (.*)"

#define APEI_ARM_MIDR	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   MIDR: (.*)"
#define APEI_ARM_MPIDR	".*\\{([0-9]+)\\}\\[Hardware Error\\]:   Multiprocessor Affinity Register \\(MPIDR\\): (.*)"

static int decompress_deflate(const char *compressed_data, ssize_t compressed_data_size,
			char *decompressed_data, ssize_t *decompressed_data_size, z_stream *zstream)
{
	int ret = Z_OK;

	ret = inflateReset2(zstream, -MAX_WBITS);
	if (ret != Z_OK)
		return ret;

	zstream->next_in = (Bytef *)compressed_data;
	zstream->avail_in = compressed_data_size;
	zstream->next_out = (Bytef *)decompressed_data;
	zstream->avail_out = *decompressed_data_size;

	ret = inflate(zstream, Z_FINISH);
	if (ret != Z_STREAM_END)
		return Z_DATA_ERROR;

	*decompressed_data_size = zstream->total_out;

	return ret;
}

static void apei_report_mem(struct trace_seq *s, struct apei *apei)
{
	struct ras_mc_event ev = {0};
	char msg_buf[400];
	time_t t;
	struct tm *tm;

	ev.erst = 1;
	if (!apei->time)
		t = time(NULL);
	else
		t = apei->time;

	tm = localtime(&t);
	if (tm)
		strftime(ev.timestamp, sizeof(ev.timestamp),
			"%Y-%m-%d %H:%M:%S %z", tm);

	ev.error_count = 1;
	ev.grain = 1;
	ev.top_layer = -1;
	ev.middle_layer = -1;
	ev.lower_layer = -1;

	switch (apei->sev) {
	case GHES_SEV_CORRECTED:
		ev.error_type = "Corrected";
		break;
	case GHES_SEV_RECOVERABLE:
		ev.error_type = "Uncorrected";
		break;
	case GHES_SEV_PANIC:
		ev.error_type = "Fatal";
		break;
	default:
		ev.error_type = "Info";
	}
	ev.severity = apei->sev;

	snprintf(msg_buf, 400, "APEI location: %s status(0x00000000): %s",
						apei->mem.loc,
						apei->mem.status ? apei->mem.status : "");
	ev.driver_detail = msg_buf;

	ev.address = apei->mem.addr;
	ev.mc_index = 0;

#ifdef HAVE_JSON_REPORT
	report_mc_event_json(s, &ev);
#endif
}

/* bit field meaning for correctable error */
static const char *aer_cor_errors[32] = {
	/* Correctable errors */
	[0]  = "Receiver Error",
	[6]  = "Bad TLP",
	[7]  = "Bad DLLP",
	[8]  = "RELAY_NUM Rollover",
	[12] = "Replay Timer Timeout",
	[13] = "Advisory Non-Fatal",
	[14] = "Corrected Internal Error",
};

/* bit field meaning for uncorrectable error */
static const char *aer_uncor_errors[32] = {
	/* Uncorrectable errors */
	[4]  = "Data Link Protocol",
	[12] = "Poisoned TLP",
	[13] = "Flow Control Protocol",
	[14] = "Completion Timeout",
	[15] = "Completer Abort",
	[16] = "Unexpected Completion",
	[17] = "Receiver Overflow",
	[18] = "Malformed TLP",
	[19] = "ECRC",
	[20] = "Unsupported Request",
};

static void apei_report_pcie(struct trace_seq *s, struct apei *apei)
{
	struct ras_aer_event ev = {0};
	unsigned long long status_val;
	char buf[1024];
	time_t t;
	struct tm *tm;

	ev.erst = 1;
	if (!apei->time)
		t = time(NULL);
	else
		t = apei->time;

	tm = localtime(&t);
	if (tm)
		strftime(ev.timestamp, sizeof(ev.timestamp),
			"%Y-%m-%d %H:%M:%S %z", tm);

	ev.dev_name = apei->pcie.port;
	ev.vendor_id = strtoul(apei->pcie.vendor_id, NULL, 16);
	ev.device_id = strtoul(apei->pcie.device_id, NULL, 16);

	if (apei->pcie.status) {
		status_val = strtoull(apei->pcie.status, NULL, 16);

		if (apei->sev == GHES_SEV_CORRECTED)
			bitfield_msg(buf, sizeof(buf), aer_cor_errors, 32, 0, 0, status_val);
		else
			bitfield_msg(buf, sizeof(buf), aer_uncor_errors, 32, 0, 0, status_val);
	} else {
		snprintf(buf, 1024, "no status");
	}
	ev.msg = buf;

	ev.tlp_header_valid = (apei->pcie.tlp_hdr != NULL);
	if (ev.tlp_header_valid)
		snprintf((buf + strlen(ev.msg)), 1024 - strlen(ev.msg),
			" TLP Header: %s", apei->pcie.tlp_hdr);

	ev.severity = apei->sev;
	switch (apei->sev) {
	case GHES_SEV_RECOVERABLE:
		ev.error_type = "Uncorrected (Non-Fatal)";
		break;
	case GHES_SEV_PANIC:
		ev.error_type = "Uncorrected (Fatal)";
		break;
	case GHES_SEV_CORRECTED:
		ev.error_type = "Corrected";
		break;
	default:
		ev.error_type = "Unknown severity";
	}

#ifdef HAVE_JSON_REPORT
	report_aer_event_json(s, &ev);
#endif
}

static void report_apei(struct apei *apei)
{
	struct trace_seq seq;
	time_t t;
	struct tm *tm;
	char timestamp[64];

	if (!apei->type)
		return;

	trace_seq_init(&seq);
	//trace_seq_printf(&seq, "{ \"event_name\": \"%s\", ", ERST_PANIC_NAME);

	switch (apei->type) {
	case APEI_MEM:
		apei_report_mem(&seq, apei);
		break;
	case APEI_PCIE:
		apei_report_pcie(&seq, apei);
		break;
	case APEI_CPU:
		if (!apei->time)
			t = time(NULL);
		else
			t = apei->time;

		tm = localtime(&t);
		if (tm)
			strftime(timestamp, sizeof(timestamp),
				"%Y-%m-%d %H:%M:%S %z", tm);

		trace_seq_printf(&seq, "{ \"%s\": \"%s\", ", JSON_REPORT_KEY, "erst_cpu");
		trace_seq_printf(&seq, "\"timestamp\": \"%s\", ", timestamp);
		trace_seq_printf(&seq, "\"fru\": \"%s\", ", apei->fru ? apei->fru : "");
		trace_seq_printf(&seq, "\"severity\": \"%s\", ", severity_strs[apei->sev]);
		trace_seq_printf(&seq, "\"cpu_id\": \"%s\" ", apei->cpu.cpu_id ? apei->cpu.cpu_id : "");
		trace_seq_puts(&seq, "}");
		break;
	case APEI_ARM:
		if (!apei->time)
			t = time(NULL);
		else
			t = apei->time;

		tm = localtime(&t);
		if (tm)
			strftime(timestamp, sizeof(timestamp),
				"%Y-%m-%d %H:%M:%S %z", tm);

		trace_seq_printf(&seq, "{ \"%s\": \"%s\", ", JSON_REPORT_KEY, "erst_arm_cpu");
		trace_seq_printf(&seq, "\"timestamp\": \"%s\", ", timestamp);
		trace_seq_printf(&seq, "\"fru\": \"%s\", ", apei->fru ? apei->fru : "");
		trace_seq_printf(&seq, "\"severity\": \"%s\", ", severity_strs[apei->sev]);
		trace_seq_printf(&seq, "\"midr\": \"%s\" ", apei->arm.midr ? apei->arm.midr : "");
		trace_seq_printf(&seq, "\"mpidr\": \"%s\" ", apei->arm.mpidr ? apei->arm.mpidr : "");
		trace_seq_puts(&seq, "}");
		break;
	}

	//trace_seq_puts(&seq, "}");
	trace_seq_do_printf(&seq);
	printf("\n");
	fflush(stdout);
	trace_seq_destroy(&seq);

	memset(apei, 0, sizeof(*apei));
	apei->err_id = -1;
}

static int is_compressed_file(const char *name)
{
	char buf[32];

	snprintf(buf, sizeof(buf), "%s", name + strlen(name) - strlen(DMESG_ERST_SUFFIX));
	return strncmp(buf, DMESG_ERST_SUFFIX, sizeof(DMESG_ERST_SUFFIX)) == 0;
}

static int line_is_panic_part1(char *line)
{
	int count, part;

	if (sscanf(line, "Panic#%d Part%u", &count, &part) != 2)
		return 0;

	return part == 1;
}

static int compressed_file_is_panic_part1(char *buf, const char *name, z_stream *zstream)
{
	ssize_t out_size = 0;
	char out_buf[128], *line;

	if (decompress_deflate(buf, strlen(buf), out_buf, &out_size, zstream))
		return 0;

	line = strtok(out_buf, "\n");

	return line_is_panic_part1(line);
}

static int file_is_panic_part1(FILE *file, const char *name, z_stream *zstream)
{
	char line[32];

	if (!fgets(line, 32, file))
		return 0;

	if (is_compressed_file(name))
		return compressed_file_is_panic_part1(line, name, zstream);

	return line_is_panic_part1(line);

}

static void regex_group(regmatch_t *m, int i, const char *line, char *buf)
{
	int e, s;

	s = m[i].rm_so;
	e = m[i].rm_eo;
	if (s >= 0)
		snprintf(buf, e - s + 1, "%s", line + s);
	else
		buf = NULL;
}

static int dmesg_erst_line_process(const char *line, struct apei_regex *regex, struct apei *apei)
{
	int ret, err_id = 0, apei_id = 0;
	regmatch_t matches[4];
	char buf[128];
	regex_t *re;
	time_t t;

	ret = regexec(re = &regex->hdr, line, 4, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		regex_group(matches, 2, line, buf);
		apei_id = atoi(buf);

		if (apei->id && apei_id != apei->id)
			report_apei(apei);
		apei->id = apei_id;

		regex_group(matches, 1, line, buf);
		t = atoll(buf);

		if (last_reboot_time)
			apei->time = last_reboot_time + t;
		else
			apei->time = 0;

		return 0;
	}

	ret = regexec(re = &regex->error, line, 4, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		regex_group(matches, 2, line, buf);
		err_id = atoi(buf);

		if (apei->err_id != -1 && err_id != apei->err_id)
			report_apei(apei);

		apei->err_id = err_id;

		regex_group(matches, 3, line, buf);
		if (!strcmp("corrected", buf))
			apei->sev = GHES_SEV_CORRECTED;
		else if (!strcmp("recoverable", buf))
			apei->sev = GHES_SEV_RECOVERABLE;
		else if (!strcmp("fatal", buf))
			apei->sev = GHES_SEV_PANIC;
		else
			apei->sev = GHES_SEV_NO;
		return 0;
	}

	if (!apei->type) {
		ret = regexec(re = &regex->type, line, 4, matches, 0);
		if (ret)
			goto error;

		regex_group(matches, 2, line, buf);
		if (!strcmp("general processor error", buf))
			apei->type = APEI_CPU;
		else if (!strcmp("memory error", buf))
			apei->type = APEI_MEM;
		else if (!strcmp("PCIe error", buf))
			apei->type = APEI_PCIE;
		else if (!strcmp("ARM processor error", buf))
			apei->type = APEI_ARM;
		else
			apei->type = APEI_NONE;

		return 0;
	}

	switch (apei->type) {
	case APEI_CPU:
		ret = regexec(re = &regex->cpu_id, line, 4, matches, 0);
		if (ret)
			goto error;
		regex_group(matches, 2, line, buf);
		apei->cpu.cpu_id = strdup(buf);

		return 0;
	case APEI_ARM:
		if (!apei->arm.midr) {
			ret = regexec(re = &regex->midr, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->arm.midr = strdup(buf);

				return 0;
			}
		}

		if (!apei->arm.mpidr) {
			ret = regexec(re = &regex->mpidr, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->arm.mpidr = strdup(buf);

				return 0;
			}
		}

		return 0;
	case APEI_MEM:
		if (!apei->mem.addr) {
			ret = regexec(re = &regex->addr, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->mem.addr = strtoull(buf, NULL, 16);
				return 0;
			}
		}

		if (!apei->mem.loc) {
			ret = regexec(re = &regex->loc, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->mem.loc = strdup(buf);
				return 0;
			}
		}

		if (!apei->mem.type) {
			ret = regexec(re = &regex->mem_type, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->mem.type = strdup(buf);
				return 0;
			}
		}

		if (!apei->mem.status) {
			ret = regexec(re = &regex->mem_status, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->mem.status = strdup(buf);
				return 0;
			}
		}

	case APEI_PCIE:
		//port type
		ret = regexec(re = &regex->port_type, line, 4, matches, 0);
		if (ret > REG_NOMATCH) {
			goto error;
		} else if (!ret) {
			regex_group(matches, 2, line, buf);
			apei->pcie.port_type = atoi(buf);

			return 0;
		}

		// port
		if (!apei->pcie.port) {
			ret = regexec(re = &regex->port, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->pcie.port = strdup(buf);
				return 0;
			}
		}

		// vendor id device id
		if (!apei->pcie.vendor_id) {
			ret = regexec(&regex->id, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->pcie.vendor_id = strdup(buf);
				regex_group(matches, 3, line, buf);
				apei->pcie.device_id = strdup(buf);

				return 0;
			}
		}

		// status
		if (!apei->pcie.status) {
			ret = regexec(re = &regex->status, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->pcie.status = strdup(buf);
				regex_group(matches, 3, line, buf);
				apei->pcie.mask = strdup(buf);

				return 0;
			}
		}

		// aer sev
		if (!apei->pcie.sev) {
			ret = regexec(re = &regex->aer_sev, line, 4, matches, 0);
			if (ret > REG_NOMATCH) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->pcie.sev = strdup(buf);

				return 0;
			}
		}

		// tlp hdr
		if (!apei->pcie.tlp_hdr) {
			ret = regexec(re = &regex->tlp_hdr, line, 4, matches, 0);
			if (ret) {
				goto error;
			} else if (!ret) {
				regex_group(matches, 2, line, buf);
				apei->pcie.tlp_hdr = strdup(buf);

				return 0;
			}
		}
	}

error:
	if (ret == REG_NOMATCH)
		return 0;
	regerror(ret, re, buf, sizeof(buf));
	printf("Regex execution error: %s\n", buf);
	return ret;
}

static int handle_erst_dmesg(FILE *file, const char *name, z_stream *zstream, struct apei_regex *regex)
{
	long fileSize;
	char *file_buf, *line, *out_data = NULL;
	ssize_t out_max_size, out_data_size = 0, bytesRead;
	int ret = 0, line_number = 1;
	struct apei apei = {0};

	apei.err_id = -1;

	if (!file_is_panic_part1(file, name, zstream))
		return -1;

	if (fseek(file, 0, SEEK_END) != 0)
		return -1;

	fileSize = ftell(file);
	if (fileSize == -1)
		return -1;

	file_buf = (char *)malloc(fileSize + 1);
	if (!file_buf)
		return -1;

	rewind(file);
	bytesRead = fread(file_buf, 1, fileSize, file);
	if (bytesRead != fileSize) {
		ret = -1;
		goto free_file;
	}
	file_buf[fileSize] = '\0';

	if (is_compressed_file(name)) {
		out_max_size = fileSize * 3;
		out_data = (char *)malloc(out_max_size);
		if (!out_data) {
			ret = -1;
			goto free_file;
		}

		ret = decompress_deflate(file_buf, fileSize, out_data, &out_data_size, zstream);
		if (ret)
			goto free_out;

		file_buf = out_data;
	}

	line = strtok(file_buf, "\n");

	while (line) {
		dmesg_erst_line_process(line, regex, &apei);

		line = strtok(NULL, "\n");
		line_number++;
	}

	report_apei(&apei);

free_out:
	if (out_data)
		free(out_data);
free_file:
	free(file_buf);

	return ret;
}

static int init_reg(regex_t *re, const char *str)
{
	char buf[128];
	int ret = 0;

	ret = regcomp(re, str, REG_EXTENDED);
	if (ret) {
		regerror(ret, re, buf, sizeof(buf));
		printf("Regex execution error: %s\n", buf);
		return ret;
	}

	return ret;
}

static void handle_erst_dmesg_file(const char *dir_name, const char *d_name, z_stream *zstream, struct apei_regex *regex)
{
	char file_path[512];
	FILE *file;

	if (strncmp(d_name, DMESG_ERST_PREFIX, strlen(DMESG_ERST_PREFIX)))
		return;

	snprintf(file_path, sizeof(file_path), "%s/%s", dir_name, d_name);

	file = fopen(file_path, "r");
	if (!file) {
		log(ALL, LOG_INFO, "Failed to open file %s\n", file_path);
		return;
	}

	handle_erst_dmesg(file, file_path, zstream, regex);

	fclose(file);

	if (erst_delete && unlink(file_path)) {
		log(ALL, LOG_INFO, "Error deleting file %s\n", file_path);
		return;
	}
}

void handle_erst_panic(void)
{
	z_stream zstream = { 0 };
	int rc = 0;
	struct dirent *entry;
	struct apei_regex regex;

	if (!last_reboot_time)
		get_last_reboot_time();

	if (init_reg(&regex.hdr, APEI_HEADER) ||
		init_reg(&regex.severity, APEI_SEVERITY) ||
		init_reg(&regex.error, APEI_ERROR) ||
		init_reg(&regex.fru, APEI_MEM_FRU) ||
		init_reg(&regex.type, APEI_TYPE) ||
		init_reg(&regex.addr, APEI_MEM_ADDR) ||
		init_reg(&regex.loc, APEI_MEM_LOC) ||
		init_reg(&regex.mem_type, APEI_MEM_TYPE) ||
		init_reg(&regex.mem_status, APEI_MEM_STATUS) ||
		init_reg(&regex.port_type, APEI_PORT_TYPE) ||
		init_reg(&regex.port, APEI_PORT) ||
		init_reg(&regex.id, APEI_ID) ||
		init_reg(&regex.status, APEI_STATUS) ||
		init_reg(&regex.aer_sev, APEI_AER_SEVE) ||
		init_reg(&regex.tlp_hdr, APEI_TLP_HDR) ||
		init_reg(&regex.cpu_id, APEI_CPU_ID) ||
		init_reg(&regex.midr, APEI_ARM_MIDR) ||
		init_reg(&regex.mpidr, APEI_ARM_MPIDR))
		return;

	DIR *dir = opendir(ERST_PATH);

	if (!dir) {
		log(ALL, LOG_INFO, "%s Failed to open directory %s\n", ERST_PATH, strerror(errno));
		return;
	}

	inflateInit2(&zstream, -MAX_WBITS);
	if (rc != Z_OK) {
		log(ALL, LOG_INFO, "Failed to open init inflate %d\n", rc);
		return;
	}

	while ((entry = readdir(dir)) != NULL) {
		struct stat path_stat;
		char file_path[MAX_PATH];

		snprintf(file_path, sizeof(file_path), "%s/%s", ERST_PATH, entry->d_name);
		stat(file_path, &path_stat);

		if (S_ISDIR(path_stat.st_mode) && !strncmp("erst", entry->d_name, sizeof("erst"))) {
			DIR *subdir = opendir(file_path);
			struct dirent *subentry;

			if (!subdir) {
				log(ALL, LOG_INFO, "Failed to open directory %s\n", strerror(errno));
				break;
			}
			while ((subentry = readdir(subdir)) != NULL)
				handle_erst_dmesg_file(file_path, subentry->d_name, &zstream, &regex);

			closedir(subdir);

		} else
			handle_erst_dmesg_file(ERST_PATH, entry->d_name, &zstream, &regex);
	}

	closedir(dir);

	inflateEnd(&zstream);
}
