#include "ras-time.h"
#define _GNU_SOURCE
#include <regex.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/syslog.h>
#include <sys/time.h>
#include <traceevent/event-parse.h>
#include "ras-logger.h"
#include "ras-report.h"

#include "ras-kmsg.h"
#include "trigger.h"

#define AMDGPU_ERROR_HEADER	".*amdgpu (.*): \\{[0-9]+\\}\\[Hardware Error\\]: Accelerator Check Architecture events logged\n"
#define AMDGPU_ERROR_STATUS	".*amdgpu (.*): \\{[0-9]+\\}\\[Hardware Error\\]: aca entry\\[[0-9]+\\].STATUS=(0x[0-9A-Fa-f]+)\n"
#define AMDGPU_ERROR_ADDR	".*amdgpu (.*): \\{[0-9]+\\}\\[Hardware Error\\]: aca entry\\[[0-9]+\\].ADDR=(0x[0-9A-Fa-f]+)\n"
#define AMDGPU_ERROR_MISC0	".*amdgpu (.*): \\{[0-9]+\\}\\[Hardware Error\\]: aca entry\\[[0-9]+\\].MISC0=(0x[0-9A-Fa-f]+)\n"
#define AMDGPU_ERROR_IPID	".*amdgpu (.*): \\{[0-9]+\\}\\[Hardware Error\\]: aca entry\\[[0-9]+\\].IPID=(0x[0-9A-Fa-f]+)\n"
#define AMDGPU_ERROR_SYND	".*amdgpu (.*): \\{[0-9]+\\}\\[Hardware Error\\]: aca entry\\[[0-9]+\\].SYND=(0x[0-9A-Fa-f]+)\n"

#define AMDGPU_MCA_ENABLED	"AMDGPU_MCA_ENABLED"

static struct amdgpu_tracer *amdgpu_tracer;
static struct amdgpu_error *amdgpu_error;
static int amdgpu_mca_enable;

static void report_amdgpu_mca(struct amdgpu_error *e)
{
#ifdef HAVE_MCE
	struct ras_mc_offline_event event;

	event.smca = true;
	event.family = 0x17;
	event.model = 0x17;
	event.bank = 1;
	event.status = e->status;
	event.synd = e->synd;
	event.ipid = e->ipid;
	event.addr = e->addr;
	event.misc0 = e->misc0;
	event.domain = e->seq;
	event.bus = e->bus;
	event.device = e->dev;
	event.function = e->func;

	ras_offline_mce_event(&event);
#endif
}

static void report_amdgpu_error(struct amdgpu_error *e)
{
	if (amdgpu_mca_enable && e->ipid && e->status)
		report_amdgpu_mca(e);
	else
		report_amdgpu_error_json(e);
}

static void regex_group(regmatch_t *m, int i, const char *line, char *buf)
{
	int e, s;

	s = m[i].rm_so;
	e = m[i].rm_eo;
	if (s >= 0)
		snprintf(buf, e - s + 1, "%s", line + s);
	else
		buf = NULL;
}

void amdgpu_tracer_match(char *msg)
{
	regmatch_t matches[10];
	regex_t *re;
	char buf[128];
	int ret;

	ret = regexec(re = &(amdgpu_tracer->header), msg, 2, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		if (amdgpu_error->tracing) {
			report_amdgpu_error(amdgpu_error);
		}

		memset(amdgpu_error, 0, sizeof(*amdgpu_error));
		amdgpu_error->tracing = 1;

		get_kmsg_time(msg, amdgpu_error->timestamp);

		regex_group(matches, 1, msg, buf);
		sscanf(buf, "%x:%x:%x.%x",
		       &amdgpu_error->seq,
		       &amdgpu_error->bus,
		       &amdgpu_error->dev,
		       &amdgpu_error->func);

		return;
	}

	ret = regexec(re = &amdgpu_tracer->status, msg, 3, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		regex_group(matches, 2, msg, buf);
		amdgpu_error->status = strtoull(buf, NULL, 16);

		return;
	}

	ret = regexec(re = &amdgpu_tracer->addr, msg, 3, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		regex_group(matches, 2, msg, buf);
		amdgpu_error->addr = strtoull(buf, NULL, 16);

		return;
	}

	ret = regexec(re = &amdgpu_tracer->misc0, msg, 3, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		regex_group(matches, 2, msg, buf);
		amdgpu_error->misc0 = strtoull(buf, NULL, 16);

		return;
	}

	ret = regexec(re = &amdgpu_tracer->ipid, msg, 3, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		regex_group(matches, 2, msg, buf);
		amdgpu_error->ipid = strtoull(buf, NULL, 16);

		return;
	}

	ret = regexec(re = &amdgpu_tracer->synd, msg, 3, matches, 0);
	if (ret > REG_NOMATCH) {
		goto error;
	} else if (!ret) {
		regex_group(matches, 2, msg, buf);
		amdgpu_error->synd = strtoull(buf, NULL, 16);

		report_amdgpu_error(amdgpu_error);
		amdgpu_error->tracing = 0;

		return;
	}

error:
	if (ret == REG_NOMATCH)
		return;
	regerror(ret, re, buf, sizeof(buf));
	printf("Regex execution error: %s\n", buf);
	return;
}

int amdgpu_tracer_destroy(void)
{
	log(ALL, LOG_INFO, "amdgpu tracer destroy\n");

	if (!amdgpu_error)
		free(amdgpu_error);

	if (!amdgpu_tracer)
		free(amdgpu_tracer);

	return 0;
}

static int init_reg(regex_t *re, const char *str)
{
	char buf[128];
	int ret = 0;

	ret = regcomp(re, str, REG_EXTENDED);
	if (ret) {
		regerror(ret, re, buf, sizeof(buf));
		printf("Regex execution error: %s\n", buf);
		return ret;
	}

	return ret;
}

int amdgpu_tracer_init(void)
{
	char *s;

	s = getenv(AMDGPU_MCA_ENABLED);
	if (!s || strcmp(s, "1"))
		amdgpu_mca_enable = 0;
	else
		amdgpu_mca_enable = 1;

	amdgpu_error = calloc(1, sizeof(struct amdgpu_error));
	if (!amdgpu_error)
		return -1;

	amdgpu_tracer = calloc(1, sizeof(struct amdgpu_tracer));
	if (!amdgpu_tracer)
		return -1;

	if (init_reg(&amdgpu_tracer->header, AMDGPU_ERROR_HEADER) ||
	    init_reg(&amdgpu_tracer->status, AMDGPU_ERROR_STATUS) ||
	    init_reg(&amdgpu_tracer->addr, AMDGPU_ERROR_ADDR) ||
	    init_reg(&amdgpu_tracer->misc0, AMDGPU_ERROR_MISC0) ||
	    init_reg(&amdgpu_tracer->ipid, AMDGPU_ERROR_IPID) ||
	    init_reg(&amdgpu_tracer->synd, AMDGPU_ERROR_SYND))
		log(ALL, LOG_ERR, "amdgpu tracer init failed\n");

	return 0;
}