#define _GNU_SOURCE
#include <regex.h>
#include <string.h>
#include <stdlib.h>

#include "ras-logger.h"
#include "ras-report.h"
#include "ras-kmsg.h"
#include "trigger.h"

int kmsg_monitor = 1;
int kmsg_trace_end;
int kmsg_limit;

struct kmsg_tracer_info *kmsg_tracer;
int kmsg_tracer_num;

int kmsg_match(char *msg)
{
	int ret, group_count, i;
	regex_t *regex;
	regmatch_t *matches;
	char tmpbuf[256];

	for (i = 0; i < kmsg_tracer_num; i++) {
		regex = &kmsg_tracer[i].regex_c;
		matches = &kmsg_tracer[i].matches[0];
		group_count = kmsg_tracer[i].group_count;

		ret = regexec(regex, msg, group_count, matches, 0);
		if (ret > REG_NOMATCH) {
			regerror(ret, regex, tmpbuf, sizeof(tmpbuf));
			log(ALL, LOG_ERR, "Regex execution error: %s\n", tmpbuf);
			return 1;
		} else if (ret == REG_NOMATCH) {
			continue;
		}

#ifdef HAVE_JSON_REPORT
		report_kmsg_event_json(&kmsg_tracer[i], msg);
#endif
		run_kmsg_trigger(kmsg_tracer, msg);

		break;
	}

	return 0;
}

int kmsg_tracer_destroy(void)
{
	log(ALL, LOG_INFO, "kmsg tracer destroy\n");

	if (!kmsg_tracer)
		return 0;
	for (int i = 0; i < kmsg_tracer_num; i++) {
		if (!kmsg_tracer[i].name)
			free(kmsg_tracer[i].name);
		if (!kmsg_tracer[i].regex)
			free(kmsg_tracer[i].regex);
		if (!kmsg_tracer[i].matches)
			free(kmsg_tracer[i].matches);
		if (!kmsg_tracer[i].group_key)
			continue;
		for (int j = 0; j < kmsg_tracer[i].group_count; j++)
			if (!kmsg_tracer[i].group_key[j])
				free(kmsg_tracer[i].group_key[j]);
			else
				continue;
		if (!kmsg_tracer[i].group_key)
			free(kmsg_tracer[i].group_key);
	}
	free(kmsg_tracer);

	amdgpu_tracer_destroy();

	return 0;
}

int kmsg_tracer_init(void)
{
	char *s;
	int kmsg_tracer_group_count, ret, c = 0;
	char buf[1026], *kmsg_tracer_name, *kmsg_tracer_regex, *tmp;
	char *kmsg_tracer_group_key, *token;

	amdgpu_tracer_init();

	s = getenv(KMSG_TRACE_END);
	if (!s)
		kmsg_trace_end = 0;
	else
		kmsg_trace_end = atoi(s);

	s = getenv(KMSG_TRACE_NUM);
	if (!s)
		return 0;

	kmsg_tracer_num = atoi(s);
	if (kmsg_tracer_num <= 0)
		return 0;

	s = getenv(KMSG_LIMIT);
	if (s) {
		kmsg_limit = atoi(s);
		if (kmsg_limit < 0)
			return -1;
	}

	kmsg_tracer = calloc(kmsg_tracer_num, sizeof(struct kmsg_tracer_info));
	if (!kmsg_tracer)
		return -1;

	for (int i = 0; i < kmsg_tracer_num; i++) {
		// trace name
		snprintf(buf, sizeof(buf), "%s_%d", KMSG_TRACER_NAME, i);
		kmsg_tracer_name = getenv(buf);
		if (!kmsg_tracer_name || ((strlen(kmsg_tracer_name) > NAME_LEN)))
			return -1;
		kmsg_tracer[i].name = strdup(kmsg_tracer_name);
		if (!kmsg_tracer[i].name)
			return -1;

		// tracer regex
		snprintf(buf, sizeof(buf), "%s_%d", KMSG_TRACER_REGEX, i);
		kmsg_tracer_regex = getenv(buf);
		if (!kmsg_tracer_regex || (strlen(kmsg_tracer_regex) > BUF_LEN))
			return -1;
		snprintf(buf, 1026, "%s\n", kmsg_tracer_regex);
		kmsg_tracer[i].regex = strdup(buf);
		if (!kmsg_tracer[i].regex)
			return -1;

		// tracer group cpunt
		snprintf(buf, sizeof(buf), "%s_%d", KMSG_TRACER_GROUP_COUNT, i);
		tmp = getenv(buf);
		if (!tmp)
			return -1;
		kmsg_tracer_group_count = atoi(tmp);
		if (kmsg_tracer_group_count < 0)
			return -1;
		kmsg_tracer_group_count++;
		kmsg_tracer[i].group_count = kmsg_tracer_group_count;
		kmsg_tracer[i].group_key = calloc(kmsg_tracer_group_count, sizeof(char *));
		if (!kmsg_tracer[i].group_key)
			return -1;

		// tracer group key
		snprintf(buf, sizeof(buf), "%s_%d", KMSG_TRACER_GROUP_KEY, i);
		kmsg_tracer_group_key = strdup(getenv(buf));
		if (!kmsg_tracer_group_key || (strlen(kmsg_tracer_group_key) > BUF_LEN))
			return -1;

		c = 0;
		token = strtok(kmsg_tracer_group_key, ",");
		while (token) {
			kmsg_tracer[i].group_key[c++] = strdup(token);
			if (c >= kmsg_tracer_group_count)
				break;
			token = strtok(NULL, ",");
		}
		free(kmsg_tracer_group_key);

		ret = regcomp(&kmsg_tracer[i].regex_c, kmsg_tracer[i].regex, REG_EXTENDED);
		if (ret) {
			regerror(ret, &kmsg_tracer[i].regex_c, buf, sizeof(buf));
			log(ALL, LOG_ERR, "Regex execution error: %s\n", buf);
			return ret;
		}

		kmsg_tracer[i].matches = calloc(kmsg_tracer_group_count, sizeof(regmatch_t));
		if (!kmsg_tracer[i].matches)
			return -1;

		if (!strcmp("xid", kmsg_tracer[i].name) ||
		    !strcmp("sxid", kmsg_tracer[i].name) ||
		    !strcmp("axid", kmsg_tracer[i].name)) {
			char *s = getenv(KMSG_IGNORE_XID);
			char *ignore;
			char *xid_token;

			if (!s)
				continue;

			ignore = strdup(s);
			if (ignore) {
				c = 0;
				xid_token = strtok(ignore, ",");
				while (xid_token) {
					kmsg_tracer[i].info.xid.ignore_xid[c++] = atoi(xid_token);
					if (c >= 30) {
						free(ignore);
						continue;
					}
					xid_token = strtok(NULL, ",");
				}
				kmsg_tracer[i].info.xid.len = c;
			}

			free(ignore);
		}
	}

	setup_event_trigger("kmsg_monitor");

	return 0;
}
