/*
 * Copyright (c) 2016, The Linux Foundation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <pci/pci.h>

#include "traceevent/event-parse.h"
#include "ras-kmsg.h"
#include "ras-report.h"
#include "ras-time.h"

#define NONE ""
int json_report = 1;

void report_mc_event_json(struct trace_seq *s, struct ras_mc_event *ev)
{
	if (!s || !ev || !json_report)
		return;

	trace_seq_printf(s,
			 "\n{ \"%s\": \"%s\", "
			 "\"timestamp\": \"%s\", "
			 "\"severity\": \"%s\", "
			 "\"error_count\": %d, "
			 "\"error_type\": \"%s\", "
			 "\"msg\": \"%s\", "
			 "\"label\": \"%s\", "
			 "\"location\": \"%d:%d:%d:%d\", "
			 "\"address\": \"%#llx\", "
			 "\"grain\": \"%#llx\", "
			 "\"syndrome\": \"%#llx\", "
			 "\"driver_detail\": \"%s\" }",
			 JSON_REPORT_KEY,
			 ev->erst ? "erst_mc_event" : "mc_event",
			 (*ev->timestamp) ? ev->timestamp : NONE,
			 severity_strs[ev->severity],
			 ev->error_count,
			 (ev->error_type) ? ev->error_type : NONE,
			 (ev->msg) ? ev->msg : NONE,
			 (ev->label) ? ev->label : NONE,
			 ev->mc_index, ev->top_layer, ev->middle_layer, ev->lower_layer,
			 ev->address,
			 ev->grain,
			 ev->syndrome,
			 (ev->driver_detail) ? ev->driver_detail : NONE);
}

static void get_pci_dev_name(const char *bdf, char *pci_name, ssize_t len, u16 *vendor_id, u16 *device_id)
{
	struct pci_access *pacc;
	struct pci_dev *dev;
	struct pci_filter filter = {0};
	int domain, bus, device, function;

	pacc = pci_alloc();
	if (!pacc)
		return;
	pci_init(pacc);
	pci_scan_bus(pacc);

	if (!pci_name)
		goto free;

	if (sscanf(bdf, "%x:%x.%x", &bus, &device, &function) == 3)
		domain = 0;
	else if (sscanf(bdf, "%x:%x:%x", &domain, &bus, &device) == 3)
		function = 0;
	else if (sscanf(bdf, "%x:%x:%x.%x", &domain, &bus, &device, &function) != 4)
		goto free;

	pci_filter_init(pacc, &filter);
	filter.bus = bus;
	filter.slot = device;
	filter.func = function;
	filter.domain = domain;

	for (dev = pacc->devices; dev; dev = dev->next) {
		if (pci_filter_match(&filter, dev)) {
			pci_fill_info(dev, PCI_FILL_IDENT);
			*vendor_id = dev->vendor_id;
			*device_id = dev->device_id;
			pci_lookup_name(pacc, pci_name, len,
					PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
					dev->vendor_id, dev->device_id);
			break;
		}
	}

free:
	pci_cleanup(pacc);
}

void report_aer_event_json(struct trace_seq *s, struct ras_aer_event *ev)
{
	char pci_name[128];
	u16 vendor = 0, device = 0;

	if (!s || !ev || !json_report)
		return;

	get_pci_dev_name(ev->dev_name, pci_name, 128, &vendor, &device);

	trace_seq_printf(s,
			 "\n{ \"%s\": \"%s\", "	\
			 "\"timestamp\": \"%s\", "	\
			 "\"severity\": \"%s\", "	\
			 "\"error_type\": \"%s\", "	\
			 "\"dev_name\": \"%s\", "	\
			 "\"pci_dev_name\": \"%s\", "	\
			 "\"vendor_id\": \"%#x\", "	\
			 "\"device_id\": \"%#x\", "	\
			 "\"msg\": \"%s\" }",
			 JSON_REPORT_KEY,
			 ev->erst ? "erst_aer_event" : "aer_event",
			 (*ev->timestamp) ? ev->timestamp : NONE,
			 severity_strs[ev->severity],
			 (ev->error_type) ? ev->error_type : NONE,
			 (ev->dev_name) ? ev->dev_name : NONE,
			 (*pci_name) ? pci_name : NONE,
			 ev->vendor_id ? ev->vendor_id : vendor,
			 ev->device_id ? ev->device_id : device,
			 (ev->msg) ? ev->msg : NONE);
}

void report_arm_event_json(struct trace_seq *s, struct ras_arm_event *ev)
{
	if (!s || !ev || !json_report)
		return;

	trace_seq_printf(s,
			 "\n{ \"%s\": \"%s\", "	\
			 "\"timestamp\": \"%s\", "	\
			 "\"error_count\": %d, "	\
			 "\"affinity\": %d, "		\
			 "\"mpidr\": \"%#lx\", "	\
			 "\"midr\": \"%#lx\", "		\
			 "\"running_state\": %d, "	\
			 "\"psci_state\": %d }",
			 JSON_REPORT_KEY,
			 ev->erst ? "erst_arm_event" : "arm_event",
			 (*ev->timestamp) ? ev->timestamp : NONE,
			 ev->error_count,
			 ev->affinity,
			 ev->mpidr,
			 ev->midr,
			 ev->running_state,
			 ev->psci_state);
}

void report_mf_event_json(struct trace_seq *s, struct ras_mf_event *ev)
{
	if (!s || !ev || !json_report)
		return;

	trace_seq_printf(s,
			 "\n{ \"%s\": \"mf_event\", \"timestamp\": \"%s\", "
			 "\"pfn\": %s, \"page_type\": \"%s\", "
			 "\"action_result\": \"%s\" }",
			 JSON_REPORT_KEY,
			 (*ev->timestamp) ? ev->timestamp : NONE,
			 (*ev->pfn) ? ev->pfn : NONE,
			 (ev->page_type) ? ev->page_type : NONE,
			 (ev->action_result) ? ev->action_result : NONE);
}

void report_signal_event_json(struct trace_seq *s, struct ras_signal_event *ev)
{
	if (!s || !ev || !json_report)
		return;

	trace_seq_printf(s,
			 "\n{ \"%s\": \"signal_event\", \"timestamp\": \"%s\", " \
			 "\"signo\": %d, \"sigerr\": %d, " \
			 "\"sigcode\": %d, \"comm\": \"%s\", " \
			 "\"pid\": %d, \"group\": %d, " \
			 "\"result\": %d }",
			 JSON_REPORT_KEY,
			 (*ev->timestamp) ? ev->timestamp : NONE,
			 ev->sig, ev->error_no, ev->code,
			 (ev->comm) ? ev->comm : NONE,
			 ev->pid, ev->group, ev->result);
}

void report_mce_event_json(struct trace_seq *s, struct mce_event *ev)
{
	if (!s || !ev || !json_report)
		return;

	if (ev->status & MCI_STATUS_UC)
		ev->severity = GHES_SEV_RECOVERABLE;
	else if (ev->status & MCI_STATUS_DEFERRED)
		ev->severity = GHES_SEV_RECOVERABLE;
	else
		ev->severity = GHES_SEV_CORRECTED;

	trace_seq_printf(s,
			 "\n{ \"%s\": \"%s\", "
			 "\"timestamp\": \"%s\", "
			 "\"severity\": \"%s\", "
			 "\"bank\": %d, "
			 "\"bank_name\": \"%s\", "
			 "\"status\": \"%#lx\", "
			 "\"error_msg\": \"%s\", "
			 "\"mcistatus_msg\": \"%s\", "
			 "\"mcastatus_msg\": \"%s\", "
			 "\"user_action\": \"%s\", "
			 "\"mc_location\": \"%s\", "
			 "\"cpuid\": \"%#x\", "
			 "\"cpu\": %d, "
			 "\"socketid\": %d, "
			 "\"ip\": \"%#lx\", "
			 "\"cs\": \"%#x\", "
			 "\"misc\": \"%#lx\", "
			 "\"addr\": \"%#lx\", "
			 "\"synd\": \"%#lx\", "
			 "\"ipid\": \"%#lx\", "
			 "\"mcgstatus_msg\": \"%s\", "
			 "\"mcgstatus\": \"%#lx\", "
			 "\"mcgcap\": \"%#lx\", "
			 "\"apicid\": \"%#x\" }",
			 JSON_REPORT_KEY,
			 ev->erst ? "erst_mce_record" : "mce_record",
			 (*ev->timestamp) ? ev->timestamp : NONE,
			 severity_strs[ev->severity],
			 ev->bank,
			 (*ev->bank_name) ? ev->bank_name : NONE,
			 ev->status,
			 (*ev->error_msg) ? ev->error_msg : NONE,
			 (*ev->mcistatus_msg) ? ev->mcistatus_msg : NONE,
			 (*ev->mcastatus_msg) ? ev->mcastatus_msg : NONE,
			 (*ev->user_action) ? ev->user_action : NONE,
			 (*ev->mc_location) ? ev->mc_location : NONE,
			 ev->cpuid,
			 ev->cpu,
			 ev->socketid,
			 ev->ip,
			 ev->cs,
			 ev->misc,
			 ev->addr,
			 ev->synd,
			 ev->ipid,
			 (*ev->mcgstatus_msg) ? ev->mcgstatus_msg : NONE,
			 ev->mcgstatus,
			 ev->mcgcap,
			 ev->apicid);
}

#ifdef HAVE_KMSG_MONITOR
void report_kmsg_event_json(struct kmsg_tracer_info *kmsg_tracer, const char *msg)
{
	struct trace_seq seq;
	int e, s;
	int group_count = kmsg_tracer->group_count;
	regmatch_t *matches = kmsg_tracer->matches;
	char tmpbuf[256] = {0}, timestamp[64] = {0};
	char pci_name[128] = {0}, *key;
	u16 vendor, device;

	get_kmsg_time(msg, timestamp);

	trace_seq_init(&seq);
	trace_seq_printf(&seq, "\n{ \"%s\": \"%s\", ", JSON_REPORT_KEY, kmsg_tracer->name);
	trace_seq_printf(&seq, "\"timestamp\": \"%s\", ", timestamp);

	for (int j = 1; j < group_count; j++) {
		s = matches[j].rm_so;
		e = matches[j].rm_eo;
		key = kmsg_tracer->group_key[j - 1];

		if (s < 0)
			trace_seq_printf(&seq, "\"%s\": \"\", ", key);
		else
			trace_seq_printf(&seq, "\"%s\": \"%.*s\"%s ",
					 key,
					 (int)(e - s), msg + s,
					 (j == group_count - 1) ? "" : ",");

		if (!strcmp("pci_port", kmsg_tracer->group_key[j - 1])) {
			snprintf(tmpbuf, 128, "%.*s", (int)(e - s), msg + s);
			get_pci_dev_name(tmpbuf, pci_name, 128, &vendor, &device);
			trace_seq_printf(&seq, "\"pci_dev_name\": \"%s\", ", pci_name);
			trace_seq_printf(&seq, "\"vendor_id\": \"%#x\", ", vendor);
			trace_seq_printf(&seq, "\"device_id\": \"%#x\", ", device);
		}

		if (!strcmp("xid", key) ||
		    !strcmp("sxid", key) ||
		    !strcmp("axid", key)) {
			int xid;

			snprintf(tmpbuf, 128, "%.*s", (int)(e - s), msg + s);
			xid = (int)strtol(tmpbuf, NULL, 10);
			for (int i = 0; i < kmsg_tracer->info.xid.len; i++) {
				if (xid == kmsg_tracer->info.xid.ignore_xid[i])
					goto out;
			}
		}
	}

	trace_seq_puts(&seq, "}");
	trace_seq_do_printf(&seq);
	printf("\n");
out:
	fflush(stdout);
	trace_seq_destroy(&seq);
}
#endif

void report_mce_offline_json(struct trace_seq *s, struct mce_event *mce,
			     struct ras_mc_offline_event *e)
{
	char tmpbuf[128] = {0}, pci_name[128] = {0};
	u16 vendor, device;

	if (!s || !e || !mce || !json_report)
		return;

	snprintf(tmpbuf, 128, "%x:%x:%x.%x", e->domain, e->bus, e->device, e->function);
	get_pci_dev_name(tmpbuf, pci_name, 128, &vendor, &device);

	trace_seq_printf(s,
			 "\n{ \"%s\": \"amdgpu_ras_event\", "	\
			 "\"timestamp\": \"%s\", "		\
			 "\"bank_name\": \"%s\", "			\
			 "\"bank\": %d, "			\
			 "\"mcastatus_msg\": \"%s\", "			\
			 "\"mcistatus_msg\": \"%s\", "			\
			 "\"mc_location\": \"%s\", "			\
			 "\"error_msg\": \"%s\", "			\
			 "\"pci_bdf\": \"%s\", "			\
			 "\"pci_dev_name\": \"%s\", "			\
			 "\"vendor_id\": \"%#x\", "			\
			 "\"device_id\": \"%#x\", "			\
			 "\"status\": \"%#lx\", "				\
			 "\"addr\": \"%#lx\", "				\
			 "\"misc0\": \"%#lx\", "				\
			 "\"ipid\": \"%#lx\", "				\
			 "\"synd\": \"%#lx\" }\n",
			 JSON_REPORT_KEY,
			 (*mce->timestamp) ? mce->timestamp : NONE,
			 (*mce->bank_name) ? mce->bank_name : NONE,
			 mce->bank,
			 (*mce->mcastatus_msg) ? mce->mcastatus_msg : NONE,
			 (*mce->mcistatus_msg) ? mce->mcistatus_msg : NONE,
			 (*mce->mc_location) ? mce->mc_location : NONE,
			 (*mce->error_msg) ? mce->error_msg : NONE,
			 tmpbuf, pci_name, vendor, device,
			 e->status, e->addr, e->misc0, e->ipid, e->synd);
}

void report_amdgpu_error_json(struct amdgpu_error *e)
{
	struct trace_seq seq;
	char tmpbuf[128] = {0}, pci_name[128] = {0};
	u16 vendor, device;

	if (!e || !json_report)
		return;

	snprintf(tmpbuf, 128, "%x:%x:%x.%x", e->seq, e->bus, e->dev, e->func);
	get_pci_dev_name(tmpbuf, pci_name, 128, &vendor, &device);

	trace_seq_init(&seq);
	trace_seq_printf(&seq,
			 "\n{ \"%s\": \"amdgpu_ras_event\", "	\
			 "\"timestamp\": \"%s\", "	\
			 "\"pci_dev_name\": \"%s\", "			\
			 "\"vendor_id\": \"%#x\", "			\
			 "\"device_id\": \"%#x\", "			\
			 "\"status\": \"0x%#lx\", "	\
			 "\"addr\": \"0x%#lx\", "		\
			 "\"misc0\": \"0x%#lx\", "	\
			 "\"ipid\": \"0x%#lx\", "		\
			 "\"synd\": \"0x%#lx\" }",
			 JSON_REPORT_KEY,
			 (*e->timestamp) ? e->timestamp : "",
			 pci_name, vendor, device,
			 e->status,
			 e->addr,
			 e->misc0,
			 e->ipid,
			 e->synd);

	trace_seq_do_printf(&seq);
	printf("\n");
	fflush(stdout);
	trace_seq_destroy(&seq);
}
