// SPDX-License-Identifier: GPL-2.0

#define _GNU_SOURCE
#include <inttypes.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#ifdef HAVE_SYSINFO
#include <sys/sysinfo.h>
#endif

#include "ras-time.h"

struct timeval boot_time;
time_t suspended_time;

int get_boot_time(struct timeval *boot_time)
{
#ifdef CLOCK_BOOTTIME
	struct timespec hires_uptime;
	struct timeval lores_uptime;
#endif
	struct timeval now;
#ifdef HAVE_SYSINFO
	struct sysinfo info;
#endif

	if (gettimeofday(&now, NULL) != 0)
		return -errno;
#ifdef CLOCK_BOOTTIME
	if (clock_gettime(CLOCK_BOOTTIME, &hires_uptime) == 0) {
		TIMESPEC_TO_TIMEVAL(&lores_uptime, &hires_uptime);
		timersub(&now, &lores_uptime, boot_time);
		return 0;
	}
#endif
#ifdef HAVE_SYSINFO
	/* fallback */
	if (sysinfo(&info) != 0)
		return -errno;

	boot_time->tv_sec = now.tv_sec - info.uptime;
	boot_time->tv_usec = 0;
	return 0;
#else
	return -ENOSYS;
#endif
}

time_t get_suspended_time(void)
{
#if defined(CLOCK_BOOTTIME) && defined(CLOCK_MONOTONIC)
	struct timespec boot, mono;

	if (clock_gettime(CLOCK_BOOTTIME, &boot) == 0 &&
	    clock_gettime(CLOCK_MONOTONIC, &mono) == 0)
		return boot.tv_sec - mono.tv_sec;
#endif
	return 0;
}

const char *skip_item(const char *begin, const char *end, const char *sep)
{
	while (begin < end) {
		int c = *begin++;

		if (c == '\0' || strchr(sep, c))
			break;
	}

	return begin;
}

void get_kmsg_time(const char *msg, char *timestamp)
{
	const char *p = msg, *end;
	char *nu = NULL;
	uint64_t usec;
	struct timeval tv = { 0 };
	time_t t;
	struct tm *tm;

	end = msg + strlen(msg) - 1;

	p = skip_item(p, end, ",");
	p = skip_item(p, end, ",;");

	errno = 0;
	usec = strtoumax(p, &nu, 10);

	if (!errno && nu && (*nu == ';' || *nu == ',')) {
		tv.tv_usec = usec % 1000000;
		tv.tv_sec = usec / 1000000;
		t = boot_time.tv_sec + suspended_time + tv.tv_sec;
	} else {
		t = time(NULL);
	}
	tm = localtime(&t);

	strftime(timestamp, 64, "%Y-%m-%d %H:%M:%S %z", tm);
}
